import numpy as np
import matplotlib as mpl
from matplotlib import pyplot as plt
from scipy.ndimage import gaussian_filter

from psytrack.aux.auxFunctions import read_input


def makeWeightPlot(wMode,
                   outData,
                   weights_dict,
                   START=0,
                   END=0,
                   perf_plot=False,
                   bias_plot=False,
                   prediction=None,
                   errorbar=None):
    """Generates a visualization of weights and related plots
    
    Displays a plot (returns nothing) of the weights over trials. Optionally,
    include a plot of bias and/or performance over time as accompanying
    subplots. Also show errorbars on weights optionally.
    
    Args:
        wMode (array): the weights to plot
        outData (dict): the dataset generating the weights
        weights_dict (dict): the weights in the dataset that were fit
        START (int): Trial to start plotting from
        END (int): Trial to stop plotting, 0 is end, negative accepted
        perf_plot (bool): Show a subplot with performance tracked over time
        bias_plot (bool): Show a subplot with choice bias tracked over time
        prediction (array OR dict): array corresponds to xval gw values from
            which to infer performance or bias. dict means infer from
            non-cross-validated weights instead. Overlaid on perf and bias plot
        errorbar (array): size of errorbars to include on each weight
    """

    ### Initialization
    K, N = wMode.shape

    if START < 0: START = N + START
    if START > N: raise Exception("START > N : " + str(START) + ", " + str(N))
    if END <= 0: END = N + END
    if END > N: END = N
    if START >= END:
        raise Exception("START >= END : " + str(START) + ", " + str(END))

    # Some useful values to have around
    maxval = np.max(np.abs(wMode)) * 1.1  # largest magnitude of any weight
    cumdays = np.cumsum(outData['dayLength'])
    myrange = np.arange(START, END)
    sigma = 20  # for smoothing performance and bias estimates

    # Custom labels for weights based on dataset
    label_names = {
        'bias': 'Bias',
        's1': 'Tone A',
        's2': 'Tone B',
        'sL': 'Left contrast',
        'sR': 'Right contrast',
        's_avg': 'Avg. Tone',
        'sBoth': 'Both contrasts',
        'h': r'Answer',
        'r': r'Reward',
        'c': r'Choice'
    }

    # Set label ordering for legend display
    label_order = {
        's1': 0,
        'sL': 0,
        's2': 1,
        'sR': 1,
        'sBoth': 1,
        'bias': 2,
        's_avg': 3,
        'h': 4,
        'c': 5,
        'r': 6
    }

    # Manually set good colors
    colors = {
        'bias': '#1982C4',
        's1': '#FF595E',
        'sL': '#FF595E',
        's2': '#FFCA3A',
        'sR': '#FFCA3A',
        's_avg': 'hotpink',
        'sBoth': 'hotpink',
        'h': '#A4036F',
        'r': '#7353BA',
        'c': '#8AC926'
    }

    # Determine species of animal being plotted for title
    if 'dataset' not in outData: plot_title = ""
    elif "Human" in outData['dataset']: plot_title = "Human "
    elif "Rat" in outData['dataset']: plot_title = "Rat "
    elif "Mouse" in outData['dataset']: plot_title = "Mouse "
    else: plot_title = ""

    ##### Plotting |
    #####----------+
    if bias_plot and perf_plot:
        fig, axs = plt.subplots(
            3,
            1,
            figsize=(12, 8),
            sharex=True,
            gridspec_kw={'height_ratios': [2, 1, 1]})
        ax = axs[0]
    elif bias_plot or perf_plot:
        fig, axs = plt.subplots(
            2,
            1,
            figsize=(12, 6),
            sharex=True,
            gridspec_kw={'height_ratios': [2, 1]})
        ax = axs[0]
    else:
        fig, ax = plt.subplots(1, 1, figsize=(12, 4))

    # Turn weights dict into weights list
    weights = []
    for i in sorted(weights_dict.keys()):
        weights += [i]*weights_dict[i]
    

    ### Top Plot, weight trajectories
    ###----------
    plt.sca(ax)

    # Plot individual weights, arrange according to order specified above
    _, order = zip(
        *sorted(zip(weights, np.arange(K)), key=lambda t: label_order[t[0]]))

    w_count = 0
    for i in order:

        if weights[i] == weights[i - 1]: w_count += 1
        else: w_count = 0
        linestyles = ['-', '--', ':', '-.']

        # Order labels and append history # if needed
        label = label_names[weights[i]]
        if weights[i] in ['h', 'r', 'c', 's_avg']:
            label += r'$^{ %s }$' % str(-(w_count + 1))

        plt.plot(
            wMode[i],
            label=label,
            lw=3,
            alpha=0.8,
            linestyle=linestyles[w_count],
            c=colors[weights[i]])

        # Plot errorbars (2SD) on weights if option is passed
        if errorbar is not None:
            plt.fill_between(
                np.arange(len(wMode[i])),
                wMode[i] - 2 * errorbar[i],
                wMode[i] + 2 * errorbar[i],
                facecolor=colors[weights[i]],
                alpha=0.2)

    # Plot vertical session lines + write text if enough space and option passed
    for i in range(len(cumdays)):
        start = cumdays[i - 1] * int(i != 0)
        plt.axvline(start, color='grey', linestyle='-', alpha=0.4)

    # Draw horizontal line at y=0
    plt.axhline(0, color='black', linestyle=':')

    # Make legend, format nicely
    (lines, labels) = plt.gca().get_legend_handles_labels()
    plt.legend(
        lines,
        labels,
        fontsize=14,
        loc='upper left',
        framealpha=0.7,
        ncol=(K + 1) // 2)

    # Add labels, adjuts other formatting
    plt.title(plot_title + outData['name'], fontsize=20)
    plt.ylabel("Weight", fontsize=18)
    plt.ylim(-maxval, maxval)
    plt.xlim(START, END + 1)
    plt.tick_params(
        axis='both', which='major', labelsize=15, direction='inout')
    ax.set_yticks(np.arange(-int(maxval), int(maxval) + 1))

    ### Test if performance or bias plot; if not, set xlabel only if this is final subplot
    xval_mask = np.ones(len(myrange)).astype(bool)

    if not (perf_plot or bias_plot):
        plt.xlabel("Trial #", fontsize=18)

    ### Data from cross-validation, arranges the inferred gw values from heldout data
    elif prediction is not None:

        X = np.array([i['gw'] for i in prediction]).flatten()
        test_inds = np.array([i['test_inds'] for i in prediction]).flatten()

        inrange = np.where((test_inds >= START) & (test_inds < END))[0]
        inds = [i for i in np.argsort(test_inds) if i in inrange]

        X = X[inds]

        untested_inds = [j for j in myrange if j not in test_inds]
        untested_inds = [np.where(myrange == i)[0][0] for i in untested_inds]
        
        xval_mask[untested_inds] = False

    # Data simply about which weights were used in the model, to reconstruct g
    else:
        g = read_input(outData, weights_dict)
        g = g[START:END]

        X = np.sum(g.T * wMode[:, START:END], axis=0)


    ### Perforamce Plot
    ###----------
    if perf_plot:
        plt.sca(axs[1])

        # Calculate smooth representation of binary accuracy, plot
        raw_correct = outData['correct'][START:END].astype(float)
        smth_correct = gaussian_filter(raw_correct, sigma=sigma)
        plt.plot(
            myrange,
            smth_correct,
            color='red',
            alpha=0.4,
            lw=3,
            linestyle='-',
            label="Empirical")

        # Calculate errorbars on empirical performance
        QQQ = np.zeros(10001)
        QQQ[5000] = 1
        QQQ = gaussian_filter(QQQ, sigma=sigma)
        perf_errorbars = np.sqrt(
            np.sum(QQQ**2) * gaussian_filter(
                (raw_correct - smth_correct)**2, sigma=sigma))
        plt.fill_between(
            myrange,
            smth_correct - 2 * perf_errorbars,
            smth_correct + 2 * perf_errorbars,
            facecolor='red',
            alpha=0.2)

        ### If prediction data is supplied, calculate the predicted accuracy
        ansR = (outData['answer'][START:END] == 2).astype(float)
        pL = 1 / (1 + np.exp(X))
        pCor = np.abs(ansR[xval_mask] - pL)
        pred_correct = gaussian_filter(pCor, sigma=sigma)
        plt.plot(
            myrange[xval_mask],
            pred_correct,
            color='maroon',
            alpha=0.8,
            lw=3,
            linestyle="-",
            label="Predicted")

        # Plot vertical session lines + write text if enough space and option passed
        for i in range(len(cumdays)):
            start = cumdays[i - 1] * int(i != 0)
            plt.axvline(start, color='grey', linestyle='-', alpha=0.4)

        # Draw horizontal line at y=0.5, random choice
        plt.axhline(0.5, color='black', linestyle=':')

        # Add labels, adjuts other formatting
        plt.ylabel("Accuracy", fontsize=18, color='red')
        plt.ylim(0.3, 1.0)
        plt.xlim(START, END + 1)
        plt.tick_params(
            axis='y',
            which='major',
            labelsize=12,
            direction='inout',
            color='red',
            labelcolor='red')
        plt.tick_params(
            axis='x',
            which='major',
            labelsize=15,
            direction='inout',
            color='black',
            labelcolor='black')
        plt.legend(fontsize=14, loc='upper left', framealpha=0.7, ncol=2)

        # Set xlabel only if this is final subplot
        if not bias_plot:
            plt.xlabel("Trial #", fontsize=18)

    ### Bias Plot
    ###----------
    if bias_plot:
        plt.sca(axs[-1])

        # Calculate smooth representation of empirical bias, plot
        choiceR = (outData['y'][START:END] == 2).astype(float)
        ansR = (outData['answer'][START:END] == 2).astype(float)
        raw_bias = choiceR - ansR
        smth_bias = gaussian_filter(raw_bias, sigma=sigma)
        plt.plot(
            myrange,
            smth_bias,
            color='blue',
            alpha=0.4,
            lw=3,
            label="Empirical")

        # Calculate errorbars on empirical performance
        QQQ = np.zeros(10001)
        QQQ[5000] = 1
        QQQ = gaussian_filter(QQQ, sigma=sigma)
        bias_errorbars = np.sqrt(
            np.sum(QQQ**2) * gaussian_filter(
                (raw_bias - smth_bias)**2, sigma=sigma))
        plt.fill_between(
            myrange,
            smth_bias - 2 * bias_errorbars,
            smth_bias + 2 * bias_errorbars,
            facecolor='blue',
            alpha=0.2)

        ### If prediction data is supplied, calculate the predicted bias
        pL = 1 / (1 + np.exp(X))
        pred_bias = (1 - pL) - ansR[xval_mask]
        smth_pred_bias = gaussian_filter(pred_bias, sigma=sigma)
        plt.plot(
            myrange[xval_mask],
            smth_pred_bias,
            color='purple',
            alpha=0.8,
            lw=3,
            label="Predicted")

        # Plot vertical session lines + write text if enough space and option passed
        for i in range(len(cumdays)):
            start = cumdays[i - 1] * int(i != 0)
            plt.axvline(start, color='grey', linestyle='-', alpha=0.4)

        # Draw horizontal line at y=0, no bias
        plt.axhline(0, color='black', linestyle=':')

        # Add labels, adjuts other formatting
        plt.ylabel("Bias", fontsize=18, color='blue')
        plt.ylim(-0.45, 0.45)  #plt.ylim(-0.6,0.6)
        plt.xlim(START, END + 1)
        plt.tick_params(
            axis='y',
            which='major',
            labelsize=12,
            direction='inout',
            color='blue',
            labelcolor='blue')
        plt.tick_params(
            axis='x',
            which='major',
            labelsize=15,
            direction='inout',
            color='black',
            labelcolor='black')
        plt.legend(fontsize=14, loc='upper left', framealpha=0.7, ncol=2)

        # Final plot, so include xlabel
        plt.xlabel("Trial #", fontsize=18)

    ### Final plotting adjustments
    ###----------
    fig.align_labels()
    plt.tight_layout(h_pad=0.1)


import numpy as np
from scipy.optimize import minimize
from scipy.sparse import csc_matrix
from scipy.sparse.linalg import spsolve

from .getMAP import getMAP, getPosteriorTerms
from psytrack.aux.auxFunctions import (
    DTinv_v,
    Dinv_v,
    DTv,
    Dv,
    DT_X_D,
    make_invSigma,
    sparse_logdet,
)


def hyperOpt(dat, hyper, weights, optList, method=None, showOpt=0, jump=2):
    """Given data and set of hyperparameters, uses decoupled Laplace to find the
    optimal hyperparameter values (i.e. the sigmas that maximize evidence)

    Args:
        dat : dict, all data from a specific subject
        hyper : dict, hyperparameters and initial values used to construct prior
            Must at least include sigma, can also include sigInit, sigDay
        weights : dict, name and count of which weights in dat['inputs'] to fit
        optList : list, hyperparameters in 'hyper' to be optimized
        method : str, control over type of fit,
            None is standard, '_days' and '_constant' also supported
        showOpt : int, 0 : no text, 1 : verbose,
            2+ : Hess + deriv check, done showOpt-1 times
        jump : int, how many times the alg can find suboptimal evd before quit

    Returns:
        best_hyper : hyperparameter values that maximizes evidence of data
        best_logEvd : log-evidence associated with optimal hyperparameter values
        best_wMode : the MAP weights found using best_hyper, maximizing logEvd
        best_Hess : dict, the Hessian found using best_hyper, maximizing logEvd
    """

    # Initialization of optimization
    opt_keywords = {
        "dat": dat,
        "hyper": hyper,
        "weights": weights,
        "optList": optList,
        "method": method,
    }

    current_hyper = hyper.copy()
    best_logEvd = None

    # Make sure all hyperparameters to be optimized are actually provided
    for val in optList:
        if (val not in hyper) or (hyper[val] is None):
            raise Exception("cannot optimize a hyperparameter not given")

    # -----
    # Hyperparameter Optimization
    # -----

    current_jump = jump
    while True:

        if best_logEvd is None:
            E0 = None
        else:
            E0 = llstruct["eMode"]  # pylint: disable=used-before-assignment

        # First get MAP for initial hyperparameter setting
        wMode, Hess, logEvd, llstruct = getMAP(
            dat,
            current_hyper,
            weights,
            E0=E0,
            method=method,
            showOpt=int(showOpt > 1))

        # Update best variables
        if best_logEvd is None:
            best_logEvd = logEvd
        if logEvd >= best_logEvd:
            current_jump = jump
            best_hyper = current_hyper.copy()
            best_logEvd = logEvd
            best_wMode = wMode
            best_Hess = Hess
        else:
            # If a worse logEvd found, reduce jump by one and
            # move hypers to midpoints, keep old bests
            current_jump -= 1
            for val in optList:
                current_hyper.update({
                    val: (current_hyper[val] + best_hyper[val]) / 2
                })

        if showOpt:
            print("\nInitial evidence:", np.round(logEvd, 5))
            for val in optList:
                print(val, np.round(np.log2(current_hyper[val]), 4))

        # Jump to end if evidence was worse enough times
        if not current_jump:
            if showOpt:
                print("Jumping to end due to no improvement in evidence")
            break

        # Now decouple prior terms from likelihood terms and store values
        K = llstruct["lT"]["ddlogli"]["K"]
        H = llstruct["lT"]["ddlogli"]["H"]
        ddlogprior = llstruct["pT"]["ddlogprior"]
        eMode = llstruct["eMode"]

        LL_v = DTinv_v(H @ Dinv_v(eMode, K), K) + ddlogprior @ eMode

        opt_keywords.update({
            "LL_terms": llstruct["lT"]["ddlogli"],
            "LL_v": LL_v
        })

        # Optimize over hyperparameters
        if showOpt:
            print("\nStarting optimization...")
            opts = {"maxiter": 15, "disp": True}
            callback = print
        else:
            opts = {"maxiter": 15, "disp": False}
            callback = None

        # Do hyperparameter optimization in log2
        optVals = []
        for val in optList:
            if np.isscalar(current_hyper[val]):
                optVals += [np.log2(current_hyper[val])]
            else:
                optVals += np.log2(current_hyper[val]).tolist()

        result = minimize(
            hyperOpt_lossfun,
            optVals,
            args=opt_keywords,
            method="BFGS",
            options=opts,
            callback=callback,
        )

        if showOpt:
            print("\nRecovered evidence:", np.round(-result.fun, 5))

        count = 0
        for val in optList:
            if np.isscalar(current_hyper[val]):
                current_hyper.update({val: 2**result.x[count]})
                count += 1
            else:
                current_hyper.update({val: 2**result.x[count:count + K]})
                count += K
            if showOpt:
                print(val, np.round(np.log2(current_hyper[val]), 4))

        # Test to see if hyperparameters have convereged
        diff = np.linalg.norm((optVals - result.x) / optVals)
        if showOpt:
            print("\nDifference:", np.round(diff, 4))
        if diff < 0.1:
            break

    # # If hyperparameters converged, recalculate evidence and wMode
    # # If evidence is not improving, no need to recalculate, jump to end
    # if current_jump:
    #     # Calculate true evidence of final set of hyperopt
    #     wMode, Hess, logEvd, llstruct = getMAP_PBups(dat,current_hyper,weights,
    #                                   method=method, showOpt=int(showOpt>1))

    #     if logEvd >= best_logEvd:
    #         best_hyper = current_hyper.copy()
    #         best_logEvd = logEvd
    #         best_wMode = wMode

    if showOpt:
        print("Coverged! Final evidence:", np.round(best_logEvd, 5))
        for val in optList:
            print(val, np.round(np.log2(best_hyper[val]), 4))

    return best_hyper, best_logEvd, best_wMode, best_Hess


def hyperOpt_lossfun(optVals, keywords):
    """Loss function used by decoupled Laplace to optimize for evidence over
    changes in hyperparameters

    Args:
        optVals : hyperparameter values currently caluclating approximate
            evidence for corresponds to hyperparameters listed in OptList
        keywords : dictionary of other values needed for optimization

    Returns:
        evd : the negative evidence (to be minimized)
    """

    # Recover N & K
    N = keywords["dat"]["y"].shape[0]
    K = keywords["LL_terms"]["K"]
    method = keywords["method"]
    dat = keywords["dat"]
    weights = keywords["weights"]

    # Reconstruct the prior covariance
    hyper = keywords["hyper"].copy()

    count = 0
    for val in keywords["optList"]:
        if np.isscalar(hyper[val]):
            hyper.update({val: 2**optVals[count]})
            count += 1
        else:
            hyper.update({val: 2**optVals[count:count + K]})
            count += K

    # Determine type of analysis (standard, constant, or day weights)
    if method is None:
        w_N = N
        # the first trial index of each new day
        days = np.cumsum(dat["dayLength"], dtype=int)[:-1]
        missing_trials = dat["missing_trials"]
    elif method == "_constant":
        w_N = 1
        days = np.array([], dtype=int)
        missing_trials = None
    elif method == "_days":
        w_N = len(dat["dayLength"])
        days = np.arange(1, w_N, dtype=int)
        missing_trials = None
    else:
        raise Exception("method " + method + " not supported")

    invSigma = make_invSigma(hyper, days, missing_trials, w_N, K)
    ddlogprior = -invSigma

    # Retrieve terms for decoupled Laplace appx.
    H = keywords["LL_terms"]["H"]
    LL_v = keywords["LL_v"]

    # Decoupled Laplace appx to new epsilon given new sigma
    DL_1 = DTv(LL_v, K)
    DL_2 = DT_X_D(ddlogprior, K)
    DL_3 = spsolve(DL_2 + H, DL_1)
    E_flat = Dv(DL_3, K)

    # Calculate likelihood and prior terms with new epsilon
    pT, lT, _ = getPosteriorTerms(
        E_flat, hyper=hyper, method=method, dat=dat, weights=weights)

    # Calculate posterior term, then approximate evidence for new sigma
    center = DL_2 + lT["ddlogli"]["H"]
    logterm_post = (1 / 2) * sparse_logdet(center)

    evd = pT["logprior"] + lT["logli"] - logterm_post

    return -evd



import numpy as np
from scipy.sparse.linalg import spsolve, inv


def jacHessCheck(fun, x0, *args, **kwargs):
    """
    Checks the accuracy of the analytic Jacobian and Hessian
    of a function that is part of the Memoize class by calculating
    finite differences in a tiny random direction.

    Args:
        fun : a function returning [result, jac, hess] that has
            been passed through a Memoize class
        x0 : the point from which to evaluate the function

    Returns:
        Nothing, simply prints out the analytic and finite
        differencing results for both the jac and hess
    """

    fun(x0, *args, **kwargs)
    JJ = fun.jacobian(x0, *args, **kwargs)
    HH = fun.hessian(x0, *args, **kwargs)

    tol = 1e-8
    randjump = np.random.random(size=x0.shape) * tol

    f1 = fun(x0 - randjump / 2, *args, **kwargs)
    JJ1 = fun.jacobian(x0 - randjump / 2, *args, **kwargs)

    f2 = fun(x0 + randjump / 2, *args, **kwargs)
    JJ2 = fun.jacobian(x0 + randjump / 2, *args, **kwargs)

    print("Analytic Jac:", np.dot(randjump, JJ))
    print("Finite Jac:  ", f2 - f1)

    if type(HH) is dict:
        print("Analytic Hess:",
              np.sum(fun.hessian_prod(x0, randjump, *args, **kwargs)))
    else:
        print("Analytic Hess:", np.sum(HH @ randjump))
    print("Finite Hess:  ", np.sum(JJ2 - JJ1))


def jacEltsCheck(fun, ind, x0, *args, **kwargs):
    """
    Checks the accuracy of individual elements in the analytic Jacobian
    of a function that is part of the Memoize class by calculating
    finite differences in the specified direction.

    Args:
        fun : a function returning [result, jac, hess] that has
            been passed through a Memoize class
        x0 : the point from which to evaluate the function

    Returns:
        Nothing, simply prints out the analytic and finite
        differencing results for the jacobian
    """

    fun(x0, *args, **kwargs)
    JJ = fun.jacobian(x0, *args, **kwargs)

    eps = 1e-5
    mask = np.zeros(len(x0))
    mask[ind] = eps

    f1 = fun(x0 - mask, *args, **kwargs)
    f2 = fun(x0 + mask, *args, **kwargs)

    dJ = (f2 - f1) / 2 / eps

    if np.sqrt((JJ[ind] - dJ)**2) > 1e-8:
        print(ind, ": ", np.sqrt((JJ[ind] - dJ)**2))
        print("Analytic Jac:", JJ[ind])
        print("Finite Jac:  ", dJ)


def hessEltsCheck(fun, ind, x0, *args, **kwargs):
    """
    Checks the accuracy of individual elements in the analytic Hessian
    of a function that is part of the Memoize class by calculating
    finite differences in the specified direction.

    Args:
        fun : a function returning [result, jac, hess] that has
            been passed through a Memoize class
        x0 : the point from which to evaluate the function

    Returns:
        Nothing, simply prints out the analytic and finite
        differencing results for the hessian
    """

    fun(x0, *args, **kwargs)
    HH = fun.hessian(x0, *args, **kwargs)

    if type(HH) is dict:
        if HH["P"].shape[0] > 1000:
            print("Hessian is too large (>1000) to check elements")
            return
        else:
            P = HH["P"]
            H = HH["H"]
            ddlogprior = HH["ddlogprior"]
            HH = -inv(P.T) @ H @ inv(P) - ddlogprior

    eps = 1e-4
    mask = np.zeros((len(x0), 2))
    mask[ind[0], 0] = eps
    mask[ind[1], 1] = eps
    v11 = fun(x0 + (mask @ np.array([[-1], [-1]]))[:, 0], *args, **kwargs)
    v21 = fun(x0 + (mask @ np.array([[1], [-1]]))[:, 0], *args, **kwargs)
    v12 = fun(x0 + (mask @ np.array([[-1], [1]]))[:, 0], *args, **kwargs)
    v22 = fun(x0 + (mask @ np.array([[1], [1]]))[:, 0], *args, **kwargs)

    dH = ((v22 - v21) - (v12 - v11)) / 4 / eps**2

    if (HH[ind[0], ind[1]] - dH)**2 > 1e-8:
        print(ind[0], ind[1], " : ", HH[ind[0], ind[1]] - dH)
        print("Analytic Hess:", HH[ind[0], ind[1]])
        print("Finite Hess:  ", dH)



import numpy as np
from datetime import datetime, timedelta
from os import makedirs
from .hyperOpt import hyperOpt


def generateSim(K=4,
                N=64000,
                hyper={},
                boundary=4.0,
                iterations=20,
                seed=None,
                savePath=None):
    """Simulates weights, in addition to inputs and multiple realizations
    of responses. Simulation data is either saved to a file or returned
    directly.

    Args:
        K : int, number of weights to simulate
        N : int, number of trials to simulate
        hyper : dict, hyperparameters and initial values used to construct the
            prior. Default is none, can include sigma, sigInit, sigDay
        boundary : float, weights are reflected from this boundary
            during simulation, is a symmetric +/- boundary
        iterations : int, # of behavioral realizations to simulate,
            same input and weights can render different choice due
            to probabilistic model, iterations are saved in 'all_Y'
        seed : int, random seed to make random simulations reproducible
        savePath : str, if given creates a folder and saves simulation data
            in a file; else data is returned

    Returns:
        save_path | (if savePath) : str, the name of the folder+file where
            simulation data was saved in the local directory
        save_dict | (if no SavePath) : dict, contains all relevant info
            from the simulation 
    """

    # Reproducability
    np.random.seed(seed)

    # Supply default hyperparameters if necessary
    sigmaDefault = 2**np.random.choice([-4.0, -5.0, -6.0, -7.0, -8.0], size=K)
    if "sigma" not in hyper:
        sigma = sigmaDefault
    elif hyper["sigma"] is None:
        sigma = sigmaDefault
    elif np.isscalar(hyper["sigma"]):
        sigma = np.array([hyper["sigma"]] * K)
    elif ((type(hyper["sigma"]) in [np.ndarray, list]) and
          (len(hyper["sigma"]) != K)):
        sigma = hyper["sigma"]
    else:
        raise Exception(
            "hyper['sigma'] must be either a scalar or a list or array of len K"
        )

    sigInitDefault = np.array([4.0] * K)
    if "sigInit" not in hyper:
        sigInit = sigInitDefault
    elif hyper["sigInit"] is None:
        sigInit = sigInitDefault
    elif np.isscalar(hyper["sigInit"]):
        sigInit = np.array([hyper["sigInit"]] * K)
    elif (type(hyper["sigInit"]) in [np.ndarray, list]) and (len(hyper["sigInit"]) != K):
        sigInit = hyper["sigInit"]
    else:
        raise Exception("hyper['sigInit'] must be either a scalar or \
            a list or array of len K")

    # sigDay not yet supported!
    if "sigDay" in hyper and hyper["sigDay"] is not None:
        raise Exception("sigDay not yet supported, please omit from hyper")

    # -------------
    # Simulation
    # -------------

    # Simulate inputs
    X = np.random.normal(size=(N, K))

    # Simulate weights
    E = np.zeros((N, K))
    E[0] = np.random.normal(scale=sigInit, size=K)
    E[1:] = np.random.normal(scale=sigma, size=(N - 1, K))
    W = np.cumsum(E, axis=0)

    # Impose a ceiling and floor boundary on W
    for i in range(len(W.T)):
        cross = (W[:, i] < -boundary) | (W[:, i] > boundary)
        while cross.any():
            ind = np.where(cross)[0][0]
            if W[ind, i] < -boundary:
                W[ind:, i] = -2 * boundary - W[ind:, i]
            else:
                W[ind:, i] = 2 * boundary - W[ind:, i]
            cross = (W[:, i] < -boundary) | (W[:, i] > boundary)

    # Save data
    save_dict = {
        "sigInit": sigInit,
        "sigma": sigma,
        "seed": seed,
        "W": W,
        "X": X,
        "K": K,
        "N": N,
    }

    # Simulate behavioral realizations in advance
    pR = 1.0 / (1.0 + np.exp(-np.sum(X * W, axis=1)))

    all_simy = []
    for i in range(iterations):
        sim_y = (pR > np.random.rand(
            len(pR))).astype(int) + 1  # 1 for L, 2 for R
        all_simy += [sim_y]

    # Update saved data to include behavior
    save_dict.update({"all_Y": all_simy})

    # Save & return file path OR return simulation data
    if savePath is not None:
        # Creates unique file name from current datetime
        folder = datetime.now().strftime("%Y%m%d_%H%M%S") + savePath
        makedirs(folder)

        fullSavePath = folder + "/sim.npz"
        np.savez_compressed(fullSavePath, save_dict=save_dict)

        return fullSavePath

    else:
        return save_dict


def recoverSim(data, N=None, iteration=0, save=False):
    """Recovers weights from the simulation data generated by generateSim()
    Can take in a filepath pointing to simulation data, or the simulation
    dict directly. Specify how many trials of data should be recovered, 
    and from which behavioral iteration (only one). Output is either saved
    in same folder as generated data, or returned directly.

    Args:
        data : str or dict, either the filepath to data from generateSim()
            or the dict returned directly by generateSim()
        N : int, number of trials to simulate, if None then just the 
            full length of the simulation
        iterations : int, which # of the behavioral realizations to recover
        save : bool, if True saves recovery data as a file in same folder
            as generateSim data (error if True, but data is not a filepath);
            if False, recovery data is returned

    Returns:
        save_path | (save=True) : str, the name of the folder+file where
            recovery data was saved in the local directory
        save_dict | (save=False) : dict, contains all relevant info
            from the recovery 
    """

    # Initialize saved recovery data
    save_dict = {"iteration": iteration}

    # Readin simulation input
    if type(data) is str:
        save_dict["simfile"] = data
        readin = np.load(data)["save_dict"].item()
    elif type(data) is dict:
        readin = data
    else:
        raise Exception("data must be either file name or dict")

    # If number of trials not specified, use all trials of simulation
    if N is None:
        N = readin["N"]
    save_dict["N"] = N

    # -------------
    # Recovery
    # -------------

    # Initialization of recovery
    K = readin["K"]
    hyper_guess = {
        # 2**-6 is an arbitrary starting point for the search
        "sigma": [2**-6] * K,
        "sigInit": readin["sigInit"],
        "sigDay": None,
    }
    optList = ["sigma"]
    weights = {"x": K}

    dat = {
        "inputs": {
            "x": readin["X"][:N, :K]
        },
        "y": readin["all_Y"][iteration][:N]
    }

    # Run recovery, recording duration of recoverty
    START = datetime.now()
    hyp, evd, wMode, _ = hyperOpt(dat, hyper_guess, weights, optList)
    END = datetime.now()

    save_dict.update({
        "K": K,
        "hyp": hyp,
        "evd": evd,
        "wMode": wMode,
        "duration": END - START
    })

    # Save (only if generateSim was also saved) or return recovery results
    if save:
        if "simfile" not in save_dict:
            raise Exception(
                "Can only save recovery if generateSim was also saved")
        save_path = (save_dict["simfile"][:-4] + "_N" + str(N) + "_i" +
                     str(iteration) + ".npz")
        np.savez_compressed(save_path, save_dict=save_dict)
        return save_path

    else:
        return save_dict



import numpy as np
from scipy.sparse.linalg import spsolve
from .auxFunctions import DT_X_D, DTinv_v, Dinv_v


class memoize(object):
    '''
    This class is a way to cache the result of a function, as well as its 1st and 2nd derivatives.
    SciPy's minimize() needs to be passed distinct objects for the jacobian + hessian, typically
    other functions that would be calculated separately. With Memoize, you can store this info
    from a function that returns [output, jac, hess] in this class object, then access the
    jacobian or hessian attribute with minimize().
    
    See the technique here: http://stackoverflow.com/a/17431749

    Args:
        a function that returns [value, jacobian, hessian]
    '''

    # Initialize class
    def __init__(self, fun):
        self.fun = fun
        self.value, self.jac, self.hess = None, None, None
        self.x = None

    # Calculate value of fun at x, along with jacobian and hessian
    def _compute(self, x, *args, **kwargs):
        self.x = np.asarray(x).copy()
        self.value, self.jac, self.hess = self.fun(x, *args, **kwargs)

    # Checks if x has changed since last call and, if so, recalculates
    # Returns value of function at x
    def __call__(self, x, *args, **kwargs):
        if self.value is not None and np.alltrue(x == self.x):
            return self.value
        else:
            self._compute(x, *args, **kwargs)
            return self.value

    # Checks if x has changed since last call and, if so, recalculates
    # Returns jacobian of function at x
    def jacobian(self, x, *args, **kwargs):
        if self.jac is not None and np.alltrue(x == self.x):
            return self.jac
        else:
            self._compute(x, *args, **kwargs)
            return self.jac

    # Checks if x has changed since last call and, if so, recalculates
    # Returns hessian of function at x (in the form of a dict)
    def hessian(self, x, *args, **kwargs):
        if self.hess is not None and np.alltrue(x == self.x):
            return self.hess
        else:
            self._compute(x, *args, **kwargs)
            return self.hess

    # Checks if x has changed since last call and, if so, recalculates
    # Returns product of hessian with arbitrary vector p (for optimization)
    def hessian_prod(self, x, p, *args, **kwargs):
        if self.hess is not None and np.alltrue(x == self.x):
            pass
        else:
            self._compute(x, *args, **kwargs)

        H = self.hess['H']
        K = self.hess['K']
        ddlogprior = self.hess['ddlogprior']

        center = DT_X_D(ddlogprior, K) + H

        return -DTinv_v(center @ Dinv_v(p, K), K)



import numpy as np
from .auxFunctions import read_input


def Kfold_crossVal(D, F=10, seed=None):
    """
    Splits a dataset into F folds, then save each individual fold
    as a test set with the other F-1 folds as a training set. Returns
    a list of F training datasets and F corresponding testing datasets

    Args:
        D : dict, data to be split into folds for cross-validation
        F : int, number of folds
        seed : int, random seed to reproduce xval fold split

    Returns:
        K_trainD : list, contains each fold's training dataset
        K_testD : list, contains each fold's testing dataset
    """

    ### Initialize randomness
    np.random.seed(seed)

    # Determine number of trials, and shuffle the order
    N = D["y"].shape[0]
    shuffled_array = np.arange(N)
    np.random.shuffle(shuffled_array)

    ### Iterate through the folds
    K_trainD = []
    K_testD = []
    for k in range(F):

        ### Define the k^th train/test split
        N_array = np.arange(N)
        chunk = int(N / F)

        ### Select the k^th chunk of shuffled trial indices
        test = np.sort(shuffled_array[k * chunk : (k + 1) * chunk])
        train = np.delete(N_array, test)

        ### Collect counts of where gaps will be in training set from missing test trials
        train_array = np.zeros(N)
        test2 = test.copy()
        while len(test2) > 0:
            train_array[test2 - 1] += 1
            test2 = np.array([i for i in test2 if i + 1 in test2])
        train_array = train_array[train]

        ### Shift any overnight gaps in test set back into training set
        if "dayLength" in D:
            day_array = np.zeros(N)
            cumDays = np.cumsum(D["dayLength"], dtype=int)[:-1]
            day_array[cumDays] = 1
            overlap = np.array([i for i in test if i in cumDays])
            while len(overlap) > 0:
                day_array[overlap + 1] = 1
                overlap = np.array([i + 1 for i in overlap if i + 1 in test])
            day_array = day_array[train]
            days = np.hstack((np.where(day_array)[0], [len(day_array)]))
            new_dayLength = np.hstack((days[0], np.diff(days)))
        else:
            new_dayLength = np.array([])

        ### Iterate through all keys in the original dict, save test/train copies
        trainD = {}
        testD = {}
        for key in D.keys():

            if key == "inputs":
                trainD[key] = {}
                testD[key] = {}
                continue

            try:
                if N == D[key].shape[0]:
                    trainD[key] = D[key][train]
                    testD[key] = D[key][test]
                else:
                    trainD[key] = D[key].copy()
                    testD[key] = D[key].copy()
            except:
                trainD[key] = D[key]
                testD[key] = D[key]

        for i in D["inputs"].keys():
            trainD["inputs"][i] = D["inputs"][i][train]
            testD["inputs"][i] = D["inputs"][i][test]

        trainD.update({"missing_trials": train_array, "dayLength": new_dayLength})
        testD.update({"test_inds": test})

        ### Append train/test dicts to list of dicts from all folds
        K_trainD += [trainD]
        K_testD += [testD]

    return K_trainD, K_testD


def Kfold_crossVal_check(testD, wMode, missing_trials, weights):
    """
    Calculates the log-likelihood and gw value of each trial in a
    test set given the wMode recovered from a corresponding training set. 

    Args:
        testD : dict, test data
        wMode : (K, trainN) array, weights recovered from training set
        missing_trials : (N*(F-1)/F,) array, indices corresponding to 
            each trial in the training set, with the value indicating
            how many test trials followed it in the original dataset
        weights : dict, name and count of weights in testD['inputs'] 
            to fit

    Returns:
        logli : array, each test trial's log-likelihood
        all_gw : array, each test trial's gw value
    """

    ### Form input matrix g from test set
    g = read_input(testD, weights)
    _, trainN = wMode.shape

    logli = []
    all_gw = []
    test_count = 0  # trial in the test set
    for t in range(trainN):  # iterate through each training trial
        
        # if training trial followed by one or more test trials
        for _ in range(int(missing_trials[t])):  

            ### Currently use the weights form the nearest prior training
            ### trial, could do interpolation...
            gw = g[test_count] @ wMode[:, t]
            yt = int(testD["y"][test_count]) - 1

            ### Save loglikelihood and gw value of each term in test set
            logli += [yt * gw - np.logaddexp(0, gw)]
            all_gw += [gw]

            ### Increment tracker of test trial index
            test_count += 1

    return np.array(logli), np.array(all_gw)



from scipy.sparse.linalg import splu
from scipy.sparse import isspmatrix_csc, diags
import numpy as np


def myblk_diags(A):
    """
	Function to take (N,K,K) matrix A and put each len(N) entry of A[:,i,j] as
	the diagonal of an (N*K x N*K) matrix. For speed, sparsity is used — define
	the diagonal values of the final matrix then use diags to construct the
	final matrix. Some minor Python black magic is used below, primarily the use
	of negative indices to index from the end, sorry.
    """
    # Retrieve shape of given matrix
    N, K, _ = np.shape(A)

    # Will need (2K-1) diagonals, with the longest N*K long
    d = np.zeros((2 * K - 1, N * K))

    # Need to keep track of each diagonal's offset in the final
    # 	matrix : (0,1,...,K-1,-K+1,...,-1)
    offsets = np.hstack((np.arange(K), np.arange(-K + 1, 0))) * N

    # Retrieve diagonal values from A to fill in d
    for i in range(K):
        for j in range(K):
            m = np.min([i, j])
            d[j - i, m * N:(m + 1) * N] = A[:, i, j]

    # After diagonals are constructed, use sparse function diags() to make
    # 	matrix, then blow up to full size
    return diags(d, offsets, shape=(N * K, N * K), format="csc")


def sparse_logdet(A):
    """
    Calculate the log determinant using sparse LU decomposition, product of 
    diagonal values of L and U matrix give determinant up to a sign change.
    Know positive determinant, so take log of the absolute value of the 
    diagonal values and sum them for log-determinant.

    See link for math and implementation details:
    http://stackoverflow.com/a/19616987  (first comment w/ Wikipedia link)
    """
    if not isspmatrix_csc(A):  # needed for splu() decomposition to work
        raise Exception(
            "sparse_logdet: matrix passed is not in sparse csc form")

    aux = splu(A)
    return np.sum(
        np.log(np.abs(aux.L.diagonal())) + np.log(np.abs(aux.U.diagonal())))


def make_invSigma(hyper, days, missing_trials, N, K):
    """Returns the banded prior matrix
    
    Constructs the random-walk prior, accounting for new sessions and trials
    omitted for cross-validation
    
    Args:
        hyper (dict): hyperparameters and values
        days (array): indices of trials that start a session
        missing_trials (array): indices of trials removed for cross-validation
        N (int): number of trials
        K (int): number of weights
    
    Returns:
        (sparse array): prior matrix
    """

    # Note: setting a sigma value at index i, adjusts the change between trials i-1 and i
    sigma = hyper["sigma"]

    if "sigInit" in hyper and hyper["sigInit"] is not None:
        sigInit = hyper["sigInit"]
    else:
        sigInit = sigma

    if "sigDay" in hyper and hyper["sigDay"] is not None:
        sigDay = hyper["sigDay"]
    else:
        sigDay = sigma

    if np.isscalar(sigma):
        invSigma_flat_k = np.ones(N) * sigma**2
        invSigma_flat_k[days] = sigDay**2  # add sigDay to day changes
        invSigma_flat_k[0] = sigInit**2  # add sigInit to beginning

        if missing_trials is not None:  # add extra sigma variance if a test trial gap
            invSigma_flat_k += missing_trials * sigma**2

        invSigma_flat = np.tile(invSigma_flat_k, K)
        return diags(invSigma_flat**-1)

    elif type(sigma) in [np.ndarray, list]:
        if len(sigma) != K:
            raise Exception("number of sigmas is not K")

        invSigma_flat = np.zeros(N * K)
        for k in range(K):
            invSigma_flat[k * N:(k + 1) * N] = sigma[k]**2

            if np.isscalar(sigDay):
                invSigma_flat[k * N + days] = sigDay**2
            else:
                invSigma_flat[k * N + days] = sigDay[k]**2

            if np.isscalar(sigInit):
                invSigma_flat[k * N] = sigInit**2
            else:
                invSigma_flat[k * N] = sigInit[k]**2

            if (missing_trials is
                    not None):  # add extra sigma variance if a test trial gap
                invSigma_flat[k * N:(k + 1) *
                              N] += missing_trials * sigma[k]**2

        return diags(invSigma_flat**-1)

    else:
        raise Exception("sigma must be of appropriate type, not" +
                        str(type(sigma)))


def read_input(D, w):
    """Creates carrier vector of inputs g
    
    From dataset D and dict of weights and counts w, collect all the inputs
    into a single K x N matrix g
    
    Args:
        D (dict): standard dataset
        w (dict): count of each type of weight to include
    
    Returns:
        g (array): matrix of inputs to model
    """

    # Determine dimensions N and K, create g
    N = len(D["y"])

    K = 0
    for i in w.keys():
        K += w[i]

    g = np.zeros((N, K))
    g_ind = 0

    for i in sorted(w.keys()):
        if i == "bias":
            g[:, g_ind:g_ind + 1] = 1
        else:
            try:
                g[:, g_ind:g_ind + w[i]] = D["inputs"][i][:, :w[i]]
            except:
                raise Exception(
                    str(i) + " given in weights not in dataset inputs")

        g_ind += w[i]

    return g


def trim(dat, START=0, END=0):
    """Utility function for slicing a dataset with a start / end point
    
    Returns a standard data set that has been sliced according to START and END
    Especially inportant for keeping session info intact
    
    Args:
        dat (dict): a standard dataset
        START (int): The trial where the new dataset should start, 0 is start
        END (int): The trial where the new dataset should end, 0 is end, can
            also take negative values
    
    Returns:
        new_dat (dict): the trimmed dataset
    """

    if (not START) and (not END):
        return dat

    N = len(dat["y"])

    if START < 0:
        START = N + START
    if START > N:
        raise Exception("START > N : " + str(START) + ", " + str(N))
    if END <= 0:
        END = N + END
    if END > N:
        END = N
    if START >= END:
        raise Exception("START >= END : " + str(START) + ", " + str(END))

    new_dat = {}
    for k in dat.keys():

        if k == "inputs":
            continue

        try:
            if N == dat[k].shape[0]:
                new_dat[k] = dat[k][START:END]
            else:
                new_dat[k] = dat[k].copy()
        except:
            new_dat[k] = dat[k]

    inputs = {}
    for i in dat["inputs"].keys():
        inputs[i] = dat["inputs"][i][START:END]
    new_dat["inputs"] = inputs

    if "dayLength" in new_dat and new_dat["dayLength"].size:
        cumdays = np.cumsum(new_dat["dayLength"])
        min_id = np.where(cumdays > START)[0][0]
        max_id = np.where(cumdays < END)[0][-1] + 1
        new = new_dat["dayLength"][min_id:max_id + 1].copy()
        new[0] = cumdays[min_id] - START
        new[-1] = END - cumdays[max_id - 1]
        if len(new) == 1:
            new[0] = END - START
        new_dat["dayLength"] = new

    new_dat["skimmed"] = {"START": START, "END": END}

    return new_dat


def DT_X_D(ddlogprior, K):
    """
    Computes D.T @ ddlogprior @ D, where D is the blocked 
    difference matrix much more quickly
    """
    dd = ddlogprior.diagonal().reshape((K, -1)).copy()

    main_diag = dd.copy()
    main_diag[:, :-1] += main_diag[:, 1:]
    main_diag = main_diag.flatten()

    off_diags = dd.copy()
    off_diags[:, 0] = 0
    off_diags = -off_diags.flatten()[1:]

    NK = main_diag.shape[0]
    A = np.zeros((3, NK))
    A[0] = main_diag
    A[1, :-1] = off_diags
    A[2, :-1] = off_diags

    return diags(A, [0, -1, 1], shape=(NK, NK), format="csc")


def Dv(v, K):
    """
    Computes D @ v, where D is the blocked difference matrix much more quickly
    """
    v2 = v.reshape(K, -1)
    v3 = np.hstack((v2[:, 0:1], np.diff(v2, axis=1)))
    v4 = v3.flatten()
    return v4


def DTv(v, K):
    """
    Computes D.T @ v, where D is the blocked difference matrix much more quickly
    """
    v2 = np.flip(v.reshape(K, -1), axis=1)
    v3 = np.hstack((v2[:, 0:1], np.diff(v2, axis=1)))
    v4 = np.flip(v3, axis=1).flatten()
    return v4


def Dinv_v(v, K):
    """
    Computes D^-1 @ v, where D is the blocked difference matrix
    much more quickly
    """
    v2 = v.reshape(K, -1)
    v3 = np.cumsum(v2, axis=1).flatten()
    return v3


def DTinv_v(v, K):
    """
    Computes D^-T @ v, where D is the blocked difference matrix much more quickly
    """
    v2 = np.flip(v.reshape(K, -1), axis=1)
    v3 = np.flip(np.cumsum(v2, axis=1), axis=1).flatten()
    return v3



import numpy as np
from scipy.sparse import csr_matrix, isspmatrix, diags, block_diag
from scipy.sparse.linalg import inv
from .auxFunctions import DT_X_D

def getCredibleInterval(Hess):
    return np.sqrt(invDiagHess(Hess)).reshape(Hess['K'],-1)

def invDiagHess(Hess):
    """
    True Hessian in e space: ddlogprior + DT^{-1} @ H @ D^{-1}
    Hessian in w space:  DT @ ddlogprior @ D + H, a block tridiagonal

    Args:
        Hess : dict, contains components needed to construct full Hessian

    Returns:
        invHess : array, the diagonal of the inverted negative Hessian
    """

    # Construct center
    center = -(DT_X_D(Hess["ddlogprior"], Hess["K"]) + Hess["H"])

    # Rearrange matrix such that it's blocked by K, not by N
    K = Hess["K"]
    N = int(Hess["ddlogprior"].shape[0] / K)
    ii = (np.reshape(np.arange(K * N), (N, -1),
                     order="F").T).flatten(order="F")
    M = center[ii]
    M = M[:, ii]

    # Do efficient inverse of block tridiagonal center
    vdiag, _, _ = invBlkTriDiag(M, K)

    # Reorder matrix back into blocked by N, rather than by K
    kk = np.argsort(ii)
    invHess = vdiag[kk]

    return invHess


def invBlkTriDiag(M, nn):
    """
    
    Efficiently inverts a block tridiagonal (a block diagonal matrix 
    with off-diagonal blocks) matrix. Blocks are (nn,nn) with nblocks
    equal to the # of blocks.
    
    Args:
        M : *sparse* square 2D array, the block tridiagonal matrix
        nn : size of each block

    Returns:
        MinvDiag : values on the diagonal of inverted M
        MinvBlocks : values on the main diagonal blocks of invM
        MinvBelowDiagBlocks : values on the off diagonal blocks of invM
    """
    if not isspmatrix(M):
        print("Casting M to sparse format")
        M = csr_matrix(M)

    nblocks = int(M.shape[0] / nn)  # number of total blocks

    # Matrices to store during recursions
    A = np.zeros((nn, nn, nblocks))  # for below-diagonal blocks
    B = np.zeros((nn, nn, nblocks))  # for diagonal blocks
    C = np.zeros((nn, nn, nblocks))  # for above-diagonal blocks
    D = np.zeros((nn, nn, nblocks))  # quantity to compute
    E = np.zeros((nn, nn, nblocks))  # quantity to compute

    # Initialize first D block
    inds = np.arange(nn)  # indices for 1st block
    B[:, :, 0] = M[np.ix_(inds, inds)].todense()
    C[:, :, 0] = M[np.ix_(inds, inds + nn)].todense()
    D[:, :, 0] = np.linalg.solve(B[:, :, 0], C[:, :, 0])

    # Initialize last E block
    inds = (nblocks - 1) * nn + inds  # indices for last block
    A[:, :, -1] = M[np.ix_(inds, inds - nn)].todense()
    B[:, :, -1] = M[np.ix_(inds, inds)].todense()
    E[:, :, -1] = np.linalg.solve(B[:, :, -1], A[:, :, -1])

    # Extract blocks A, B, and C
    for ii in np.arange(1, nblocks - 1):
        inds = np.arange(nn) + ii * nn  # indices for center block
        A[:, :, ii] = M[np.ix_(inds,
                               inds - nn)].todense()  # below-diagonal block
        B[:, :, ii] = M[np.ix_(inds, inds)].todense()  # middle diagonal block
        C[:, :, ii] = M[np.ix_(inds,
                               inds + nn)].todense()  # above diagonal block

    # Make a pass through data to compute D and E
    for ii in np.arange(1, nblocks - 1):
        # Forward recursion
        D[:, :, ii] = np.linalg.solve(
            B[:, :, ii] - A[:, :, ii] @ D[:, :, ii - 1], C[:, :, ii])

        # Backward recursion
        jj = nblocks - ii - 1
        E[:, :, jj] = np.linalg.solve(
            B[:, :, jj] - C[:, :, jj] @ E[:, :, jj + 1], A[:, :, jj])

    # Now form blocks of inverse covariance
    I = np.eye(nn)
    MinvBlocks = np.zeros((nn, nn, nblocks))
    MinvBelowDiagBlocks = np.zeros((nn, nn, nblocks - 1))
    MinvBlocks[:, :, 0] = np.linalg.inv(
        B[:, :, 0] @ (I - D[:, :, 0] @ E[:, :, 1]))
    MinvBlocks[:, :, -1] = np.linalg.inv(B[:, :, -1] -
                                         A[:, :, -1] @ D[:, :, -2])
    for ii in np.arange(1, nblocks - 1):
        # Compute diagonal blocks of inverse
        MinvBlocks[:, :, ii] = np.linalg.inv(
            (B[:, :, ii] - A[:, :, ii] @ D[:, :, ii - 1])
            @ (I - D[:, :, ii] @ E[:, :, ii + 1]))
        # Compute below-diagonal blocks
        MinvBelowDiagBlocks[:, :, ii -
                            1] = -D[:, :, ii - 1] @ MinvBlocks[:, :, ii]

    MinvBelowDiagBlocks[:, :, -1] = -D[:, :, -2] @ MinvBlocks[:, :, -1]

    # Extract just the diagonal elements
    MinvDiag = np.zeros(nn * nblocks)
    for ii in np.arange(nblocks):
        MinvDiag[ii * nn:(ii + 1) * nn] = np.diag(MinvBlocks[:, :, ii])

    return MinvDiag, MinvBlocks, MinvBelowDiagBlocks



import numpy as np
from scipy.optimize import minimize

from psytrack.aux.memoize import memoize
from psytrack.aux.jacHessCheck import jacHessCheck
from psytrack.aux.auxFunctions import (
    DT_X_D,
    sparse_logdet,
    read_input,
    make_invSigma,
    DTinv_v,
    myblk_diags,
)


def getMAP(dat, hyper, weights, method=None, E0=None, showOpt=0):
    """Estimates epsilon parameters with random walk prior

    Args:
        dat : dict, all data from a specific subject
        hyper : a dictionary of hyperparameters used to construct the prior
            Must at least include sigma, can also include sigInit, sigDay
        weights : dict, name and count of weights in dat['inputs'] to fit
        method : str, control over type of learning, defaults to standard 
            trial-by-trial fitting; '_days' and '_constant' also supported
        E0 : initial parameter estimate, must be of approprite size N*K, 
            defaults to zeros
        showOpt : {0 : no text, 1 : verbose, 
            2+ : Hess + deriv check, done showOpt-1 times}
        maxIter : max iterations of the optimizer

    Returns:
        wMode : MAP estimate of the weights
        Hess : the Hessian of the log posterior at wMode, used for Laplace appx.
            in evidence max in this case, is a dict of sparse terms needed to
            construct Hess (which is not sparse)
        logEvd : log of the evidence
        llstruct : dictionary containing the components of the log evidence and
            other info
    """

    # -----
    # Initializations and Sanity Checks
    # -----

    # Check and count trials
    if "inputs" not in dat or "y" not in dat or type(
            dat["inputs"]) is not dict:
        raise Exception("getMAP_PBups: insufficient input, missing y")
    N = len(dat["y"])

    # Check and count weights
    K = 0
    if type(weights) is not dict:
        raise Exception("weights must be a dict")
    for i in weights.keys():
        if type(weights[i]) is not int or weights[i] < 0:
            raise Exception("weight values must be non-negative ints")
        K += weights[i]

    # Check if using constant weights or by-day weights
    if method is None:
        w_N = N
    elif method == "_constant":
        w_N = 1
    elif method == "_days":
        w_N = len(dat["dayLength"])
    else:
        raise Exception("method type " + method + " not supported")

    # Initialize weights to particular values (default 0)
    if E0 is not None:
        if type(E0) is not np.ndarray:
            raise Exception("E0 must be an array")

        if E0.shape == (w_N * K,):
            eInit = E0.copy()
        elif E0.shape == (w_N, K):
            eInit = E0.flatten()
        else:
            raise Exception("E0 must be shape (w_N*K,) or (w_N,K), not " +
                            str(E0.shape))
    else:
        eInit = np.zeros(w_N * K)

    # Do sanity checks on hyperparameters
    if "sigma" not in hyper:
        raise Exception("WARNING: sigma not specified in hyper dict")
    if "alpha" in hyper:
        raise Exception("WARNING: alpha is not supported")
    if method == "_constant":
        if "sigInit" not in hyper or hyper["sigInit"] is None:
            print("WARNING: sigInit being set to sigma for method", method)
    if method == "_days":
        if "sigDay" not in hyper or hyper["sigDay"] is None:
            print("WARNING: sigDay being set to sigma for method", method)

    # Get index of start of each day
    if ("dayLength" not in dat) and (
        ("sigDay" in hyper and hyper["sigDay"] is not None) or
        (method == "_days")):
        print("WARNING: sigDay has no effect, dayLength not supplied in dat")
        dat["dayLength"] = np.array([], dtype=int)

    # Account for missing trials from running xval (i.e. gaps from test set)
    if "missing_trials" in dat and dat["missing_trials"] is not None:
        if len(dat["missing_trials"]) != N:
            raise Exception("missing_trials must be length N if used")
    else:
        dat["missing_trials"] = None

    # -----
    # MAP estimate
    # -----

    # Prepare minimization of loss function, Memoize to preserve Jac+Hess info
    lossfun = memoize(negLogPost)
    my_args = (dat, hyper, weights, method)

    if showOpt:
        opts = {"disp": True}
        callback = print
    else:
        opts = {"disp": False}
        callback = None

    # Actual optimization call
    # Uses 'hessp' to pass a function that calculates product of Hessian
    #    with arbitrary vector
    if showOpt:
        print("Obtaining MAP estimate...")
    result = minimize(
        lossfun,
        eInit,
        jac=lossfun.jacobian,
        hessp=lossfun.hessian_prod,
        method="trust-ncg",
        tol=1e-9,
        args=my_args,
        options=opts,
        callback=callback,
    )

    # Recover the results of the optimization
    eMode = result.x
    # dict of sparse components of Hess
    Hess = lossfun.hessian(eMode, *my_args)

    # Print message if optimizer does not converge (usually still pretty good)
    if showOpt and not result.success:
        print("WARNING — MAP estimate: minimize() did not converge\n",
              result.message)
        print("NOTE: this is ususally irrelevant as the optimizer still finds "
              "a good solution. If you are concerned, run a check of the "
              "Hessian by setting showOpt >= 2")

    # Run DerivCheck & HessCheck at eMode (will run ShowOpt-1 distinct times)
    if showOpt >= 2:
        print("** Jacobian and Hessian Check **")
        for check in range(showOpt - 1):
            print("\nCheck", check + 1, ":")
            jacHessCheck(lossfun, eMode, *my_args)
            print("")

    # -----
    # Evidence (Marginal likelihood)
    # -----

    # Prior and likelihood at eMode, also recovering the associated wMode
    if showOpt:
        print("Calculating evd, first prior and likelihood at eMode...")
    pT, lT, wMode = getPosteriorTerms(eMode, *my_args)

    # Posterior term (with Laplace approx), calculating sparse log determinant
    if showOpt:
        print("Now the posterior with Laplace approx...")
    center = DT_X_D(Hess["ddlogprior"], Hess["K"]) + Hess["H"]
    logterm_post = (1 / 2) * sparse_logdet(center)

    # Compute Log evd and construct dict of likelihood, prior,
    #   and posterior terms
    logEvd = lT["logli"] + pT["logprior"] - logterm_post
    if showOpt:
        print("Evidence:", logEvd)

    # Package up important terms to return
    llstruct = {"lT": lT, "pT": pT, "eMode": eMode}

    return wMode, Hess, logEvd, llstruct


def negLogPost(*args):
    """Returns negative log posterior (and its first and second derivative)
    Intermediary function to allow for getPosteriorTerms to be optimized

    Args:
        same as getPosteriorTerms()

    Returns:
        negL : negative log-likelihood of the posterior
        dL : 1st derivative of the negative log-likelihood
        ddL : 2nd derivative of the negative log-likelihood,
            kept as a dict of sparse terms!
    """

    # Get prior and likelihood terms
    [priorTerms, liTerms, _] = getPosteriorTerms(*args)  # pylint: disable=no-value-for-parameter

    # Negative log posterior
    negL = -priorTerms["logprior"] - liTerms["logli"]
    dL = -priorTerms["dlogprior"] - liTerms["dlogli"]
    ddL = {"ddlogprior": priorTerms["ddlogprior"], **liTerms["ddlogli"]}

    return negL, dL, ddL


def getPosteriorTerms(E_flat, dat, hyper, weights, method=None):
    """Given a sequence of parameters formatted as an N*K matrix, calculates
    random-walk log priors & likelihoods and their derivatives

    Args:
        E_flat : array, the N*K epsilon parameters, flattened to a single
        vector
        ** all other args are same as in getMAP **

    Returns:
        priorTerms : dict, the log-prior as well as 1st + 2nd derivatives
        liTerms : dict, the log-likelihood as well as 1st + 2nd derivatives
        W : array, the weights, calculated directly from E_flat
    """

    # !!! TEMPORARY --- Need to update !!!
    if method in ["_days", "_constant"]:
        raise Exception(
            "Need efficient calculations for _constant or _days methods")

    # ---
    # Initialization
    # ---

    # If function is called directly instead of through getMAP,
    #       fill in dummy values
    if "dayLength" not in dat:
        dat["dayLength"] = np.array([], dtype=int)
    if "missing_trials" not in dat:
        dat["missing_trials"] = None

    # Unpack input into g
    g = read_input(dat, weights)
    N, K = g.shape

    # Determine type of analysis (standard, constant, or day weights)
    if method is None:
        w_N = N
        # the first trial index of each new day
        days = np.cumsum(dat["dayLength"], dtype=int)[:-1]
        missing_trials = dat["missing_trials"]
    elif method == "_constant":
        w_N = 1
        days = np.array([], dtype=int)
        missing_trials = None
    elif method == "_days":
        w_N = len(dat["dayLength"])
        days = np.arange(1, w_N, dtype=int)
        missing_trials = None
    else:
        raise Exception("method " + method + " not supported")

    # Check shape of epsilon, with
    #   w_N (effective # of trials) * K (# of weights) elements
    if E_flat.shape != (w_N * K,):
        print(E_flat.shape, w_N, K, method)
        raise Exception("parameter dimension mismatch (#trials * #weights)")

    # ---
    # Construct random-walk prior, calculate priorTerms
    # ---

    # Construct random walk covariance matrix Sigma^-1, use sparsity for speed
    invSigma = make_invSigma(hyper, days, missing_trials, w_N, K)

    # Calculate the log-determinant of prior covariance,
    #   the log-prior, 1st, & 2nd derivatives
    logdet_invSigma = np.sum(np.log(invSigma.diagonal()))
    logprior = (1 / 2) * (logdet_invSigma - E_flat @ invSigma @ E_flat)
    dlogprior = -invSigma @ E_flat
    ddlogprior = -invSigma

    priorTerms = {
        "logprior": logprior,
        "dlogprior": dlogprior,
        "ddlogprior": ddlogprior,
    }

    # ---
    # Construct likelihood, calculate liTerms
    # ---

    # Reconstruct actual weights from E values
    E = np.reshape(E_flat, (K, w_N), order="C")
    W = np.cumsum(E, axis=1)

    # Calculate probability of Right on each trial
    y = dat["y"] - 1
    gw = np.sum(g * W.T, axis=1)
    pR = 1 / (1 + np.exp(-gw))

    # Preliminary calculations for 1st and 2nd derivatives
    dlliList = g * (y - pR)[:, None]

    alpha = (pR**2 - pR)[:, None, None]
    HlliList = alpha * (g[:, :, None] @ g[:, None, :])

    # INSERT CODE HERE TO HANDLE _days OR _constant METHODS

    # Calculate the log-likelihood and 1st & 2nd derivatives
    logli = np.sum(y * gw - np.logaddexp(0, gw))
    dlogli = DTinv_v(dlliList.flatten("F"), K)
    ddlogli = {"H": myblk_diags(HlliList), "K": K}

    liTerms = {"logli": logli, "dlogli": dlogli, "ddlogli": ddlogli}

    return priorTerms, liTerms, W



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
	name='psytrack',
	version='1.0.0',
	description='Tool for tracking dynamic psychometric curves',
	long_description=long_description,
	long_description_content_type="text/markdown",
	url='http://github.com/nicholas-roy/psytrack',
	author='Nicholas A. Roy, Ji Hyun Bak, and Jonathan W. Pillow',
	author_email='nicholas.roy.42@gmail.com',
	license='MIT',
    packages=setuptools.find_packages(),
	install_requires=[
          'numpy',
		  'scipy',
		  'matplotlib',
		  'markdown',
      ],
	classifiers=[
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
		"Intended Audience :: Science/Research",
		"Topic :: Scientific/Engineering",
		],
	zip_safe=False,
	include_package_data=True)